from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable


def register_args(make_parser: Callable[..., ArgumentParser]):
    parser = make_parser("diff")
    parser.add_argument(
        "--eval",
        type=Path,
        required=True,
        help="""
            Path to output of the `dubstub eval` command
        """,
    )
    parser.add_argument(
        "--left",
        default="pyright",
        help="""
            What to display on the left side of the diff.
        """,
    )
    parser.add_argument(
        "--right",
        default="dubstub",
        help="""
            What to display on the right side of the diff.
        """,
    )
    parser.add_argument(
        "--width",
        default=160,
        type=int,
        help="""
            Total width of tables
        """,
    )
    parser.add_argument(
        "--hide-missing",
        action="store_true",
        help="""
            Do not show diff if one of the files is entirely missing
        """,
    )
    parser.add_argument(
        "--filter",
        help="""
            Regular expression filter for the files that should be diffed
        """,
    )
    parser.set_defaults(entrypoint=main)


def main(args: Namespace):
    from .impl import run  # pylint:disable=import-outside-toplevel

    run(
        eval_path=args.eval,
        left=args.left,
        right=args.right,
        width=args.width,
        hide_missing=args.hide_missing,
        filter_re=args.filter,
    )



from dataclasses import dataclass
from enum import Enum


class Tag(Enum):
    MODULE = "module"
    CLASS = "class"
    FUNCTION = "function"
    IMPORT = "import"
    TYPE_ALIAS = "type_alias"
    VARIABLE = "variable"
    ANNOTATED = "annotated"
    ASSIGNED = "assigned"
    IF = "if"
    DOCSTRING = "docstring"
    ELLIPSIS = "ellipsis"
    SPACER = "spacer"


@dataclass
class MatchContext:
    parent_tags: set[Tag]
    tags: set[Tag]
    file_path: str
    name: str | None = None
    annotation: str | None = None
    value: str | None = None
    child_tags: set[Tag] | None = None



import re
import sys
from contextlib import contextmanager
from typing import Mapping, Sequence, TypeAlias

Json: TypeAlias = Mapping[str, "Json"] | Sequence["Json"] | str | int | float | bool | None


def regex_match(pattern: str, strings: str | list[str]) -> bool:
    if isinstance(strings, str):
        strings = [strings]

    for string in strings:
        if re.fullmatch(pattern, string):
            return True

    return False


@contextmanager
def extra_guard(extra: str, *, modules: list[str] | None = None, executables: list[str] | None = None):
    try:
        yield
    except ModuleNotFoundError as exc:
        if exc.name in (modules or []):
            print(f"The `{extra}` extra seems to not be installed (module `{exc.name}` not found)")
            sys.exit(1)
        raise
    except FileNotFoundError as exc:
        if exc.filename is None:
            raise

        if isinstance(exc.filename, bytes):
            filename = exc.filename.decode()
        else:
            filename = exc.filename

        if filename in (executables or []):
            print(f"The executable `{filename}` was not found (can be installed with `{extra}` extra)")
            sys.exit(1)

        raise



import re
from difflib import Differ
from pathlib import Path
from typing import Iterable

from rich.console import Console
from rich.table import Column, Table
from rich.text import Text

from ..fs import walk_dir


def walk_rel_stub_files(root: Path) -> Iterable[Path]:
    for path in walk_dir(root):
        if path.is_file() and (path.suffix in (".py", ".pyi") or path.name == "py.typed"):
            yield path.relative_to(root)


# pylint: disable-next=too-many-locals,too-many-branches,too-many-statements,too-many-arguments,too-many-positional-arguments
def run(
    eval_path: Path,
    left: str,
    right: str,
    width: int,
    hide_missing: bool,
    filter_re: str | None,
):
    a_path = eval_path / left
    b_path = eval_path / right

    paths: set[Path] = set()
    paths.update(walk_rel_stub_files(a_path))
    paths.update(walk_rel_stub_files(b_path))

    for path in sorted(paths):
        if filter_re is not None and not re.match(filter_re, str(path)):
            continue

        a_file = a_path / path
        b_file = b_path / path

        a_content = ""
        if a_file.exists():
            a_content = a_file.read_text()
        elif hide_missing:
            continue

        b_content = ""
        if b_file.exists():
            b_content = b_file.read_text()
        elif hide_missing:
            continue

        if a_content == b_content:
            continue

        a_lines = a_content.splitlines()
        b_lines = b_content.splitlines()

        chunks: list[tuple[str, list[str], list[str]]] = [
            ("-", [], []),
        ]

        differ = Differ()
        for line in differ.compare(a_lines, b_lines):
            marker = line[0]

            pair: tuple[str, list[str], list[str]]
            if marker == " ":
                pair = ("==", [line[2:]], [line[2:]])
            elif marker == "-":
                pair = ("!=", [line[2:]], [])
            elif marker == "+":
                pair = ("!=", [], [line[2:]])
            else:
                continue

            if chunks[-1][0] == pair[0]:
                chunks[-1][1].extend(pair[1])
                chunks[-1][2].extend(pair[2])
            else:
                chunks.append(pair)

        for ci, chunk in enumerate(chunks):
            if chunk[0] == "==":
                if len(chunk[1]) > 6 and ci + 1 < len(chunks):
                    new = [
                        *chunk[1][0:3],
                        # Text("[...]", style="bright_black"),
                        "<<SEPARATOR>>",
                        *chunk[1][-3:],
                    ]
                    chunk[1].clear()
                    chunk[2].clear()
                    chunk[1].extend(new)
                    chunk[2].extend(new)
                elif len(chunk[1]) > 3 and ci + 1 == len(chunks):
                    new = [
                        *chunk[1][0:3],
                        "<<SEPARATOR>>",
                    ]
                    chunk[1].clear()
                    chunk[2].clear()
                    chunk[1].extend(new)
                    chunk[2].extend(new)
                elif len(chunk[1]) > 3 and ci == 0:
                    new = [
                        "<<SEPARATOR>>",
                        *chunk[1][-3:],
                    ]
                    chunk[1].clear()
                    chunk[2].clear()
                    chunk[1].extend(new)
                    chunk[2].extend(new)

        console = Console(color_system="standard")
        tbl = Table(
            Column(left, width=width // 2),
            Column(right, width=width // 2),
            title=str(path),
            title_justify="left",
        )

        for kind, a_chunk_lines, b_chunk_lines in chunks:
            ml = max(len(a_chunk_lines), len(b_chunk_lines))
            for i in range(ml):
                a_chunk_line = ""
                b_chunk_line = ""

                if i < len(a_chunk_lines):
                    a_chunk_line = a_chunk_lines[i]
                if i < len(b_chunk_lines):
                    b_chunk_line = b_chunk_lines[i]

                if kind == "!=":
                    tbl.add_row(
                        Text(a_chunk_line, style="red"),
                        Text(b_chunk_line, style="green"),
                    )
                elif a_chunk_line == "<<SEPARATOR>>":
                    tbl.add_section()
                else:
                    tbl.add_row(a_chunk_line, b_chunk_line)

        console.print(tbl)
        print()



import ast
import json
import re
import subprocess
from pathlib import Path
from shutil import copy2, rmtree
from tempfile import TemporaryDirectory

from ..config import ValidatedConfig
from ..format import format_pyi_tree
from ..fs import Kind, Walker, remove, walk_dir
from ..source import AstConfig, Source


def generate_copy(inp: Path, out: Path):
    # in copy mode we just copy the pyi file as-is

    out.parent.mkdir(exist_ok=True, parents=True)
    copy2(inp, out)


def run_pyright(tmp: Path, base_path: Path, module_name: str, config: ValidatedConfig) -> Path:
    cfg = tmp / "config.json"
    cfg.write_text(
        json.dumps(
            {
                "extraPaths": [
                    str(base_path),
                ],
                # "include": [],
            }
        )
    )

    major, minor = config.get_python_version()
    subprocess.run(
        [
            "pyright",
            "--createstub",
            module_name,
            # "--verbose",
            "--pythonversion",
            f"{major}.{minor}",
            "-p",
            str(cfg),
        ],
        cwd=tmp,
        check=True,
    )

    return tmp / "typings"


def generate_pyright(inp: Path, out: Path, is_file: bool, config: ValidatedConfig):
    # otherwise we invoke pyright with the right base directory to do a src import from
    with TemporaryDirectory() as tmpdir:
        tmp = Path(tmpdir)

        # determine which module name to import from which directory
        base_path = inp.parent
        module_name = inp.stem

        # call pyright
        pyright_out = run_pyright(tmp, base_path, module_name, config)

        if is_file:
            # identify where pyright put the output, and normalize it
            gen = pyright_out / module_name / "__init__.pyi"
            normalize_out_file(gen, config)

            # copy to output
            out.parent.mkdir(parents=True, exist_ok=True)
            gen.rename(out)
        else:
            # identify where pyright put the output, and normalize it
            gen = tmp / "typings" / module_name
            for child in walk_dir(gen):
                if child.is_file() and child.suffix == ".pyi":
                    normalize_out_file(child, config)

            # copy to output
            out.parent.mkdir(parents=True, exist_ok=True)
            rmtree(out, ignore_errors=True)
            gen.rename(out)


def generate(inp_root: Path, out_root: Path, config: ValidatedConfig):
    # pylint: disable=duplicate-code

    walker = Walker(inp_root, out_root)

    for event in walker.walk():
        inp = event.inp_path
        out = event.out_path

        match event.kind:
            case Kind.ROOT:
                print(f"Clean {event.out_rel_pattern}")
                remove(out)

                if inp.is_dir() or (inp.suffix in (".py", ".pyi")):
                    print(f"Stub {event.inp_rel_pattern} -> {event.out_rel_pattern}")
                    generate_pyright(inp, out, inp.is_file(), config)
            case Kind.COPY:
                print(f"Copy {event.out_rel_pattern}")
                generate_copy(inp, out)
            case Kind.STUB:
                pass

    format_pyi_tree(walker, config)


def normalize(content: str, config: ValidatedConfig) -> str:
    """
    Remove generic docstring inserted by pyright, and restore position of existing docstring
    """

    source = Source(content, Path("<pyright output>"), AstConfig(feature_version=config.get_python_version()))
    parsed = source.parse_module()

    prefix = True
    lines: list[str] = []
    for stmt in parsed.body:
        if (
            prefix
            and isinstance(stmt, ast.Expr)
            and isinstance(stmt.value, ast.Constant)
            and isinstance(stmt.value.value, str)
        ):
            if stmt.value.value.strip() == "This type stub file was generated by pyright.":
                # skip this
                pass
            else:
                lines.insert(0, source.unparse_original_source(stmt))
                lines.insert(1, "")
                prefix = False
        elif isinstance(stmt, ast.Import | ast.ImportFrom):
            lines.append(source.unparse_original_source(stmt))
        else:
            if lines and lines[-1] != "":
                lines.append("")

            # keep the rest of the file as it is
            start, _ = source.unparse_offsets(stmt)
            for line in content[start:].splitlines():
                lines.append(line)
            break

    content = "\n".join(lines).strip() + "\n"

    content = re.sub(r" *# -> .*\n", "\n", content)

    return content


def normalize_out_file(path: Path, config: ValidatedConfig):
    content = path.read_text()
    content = normalize(content, config)
    path.write_text(content)



import ast
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Type

from ..config import ValidatedConfig
from ..config.match_ctx import MatchContext, Tag
from ..source import AstConfig, Source

CONSTANT_TYPES: list[Type[Any]] = [
    str,
    bytes,
    bool,
    int,
    float,
    complex,
    type(...),
    type(None),
]


@dataclass
class Node:
    lines: list[str] = field(default_factory=list)
    """
    Usually a node is a single line, but with annotations and
    block constructs you can end up with more than one
    """

    children: list["Node"] = field(default_factory=list)
    """
    If this node can contain sub nodes, the are collected in this list.
    These children may or may not be indented when output, depending on group tag.
    """

    tags: set[Tag] = field(default_factory=set)
    """
    Tags to identify this node
    """

    meta: dict[str, Any] = field(default_factory=dict)
    """
    Extra metadata specific to the type of node
    """

    def add_child(self, *tags: Tag) -> "Node":
        node = Node(tags=set(tags))
        self.children.append(node)
        return node

    def add_line(self, line: str, allow_multiline: bool = False):
        if not allow_multiline:
            assert "\n" not in line, line
        self.lines.append(line)

    def add_tag(self, tag: Tag):
        self.tags.add(tag)


class Logger:
    debug: bool = False

    def ignore_unhandled(self, msg: str):
        """syntax was ignored because we can not handle it yet"""

        print("Unexpected syntax ignored:", msg)

    def ignore_disabled(self, msg: str):
        """syntax was ignored because we have a config that disabled it"""

        if self.debug:
            print(msg)

    def ignore_intentional(self, msg: str):
        """syntax was intentionally ignored because it has no influence on type stubs"""

        if self.debug:
            print("Expected syntax ignored:", msg)


# pylint: disable-next=too-many-public-methods
class Stubber:
    source: Source
    config: ValidatedConfig
    used_names: set[str] | None
    logger: Logger

    def __init__(self, source: Source, config: ValidatedConfig, used_names: set[str] | None = None):
        self.source = source
        self.config = config
        self.used_names = used_names
        self.logger = Logger()

    def output(self, ast_module: ast.Module) -> str:
        # to get a more homogenous code structure, we pretend there is a parent for
        # a module ast node
        out = Node()
        self.stub(out, ast_module)
        assert len(out.children) == 1

        # fully stubbed module
        out = out.children[0]

        # render everything to a flat list of lines
        lines: list[str] = []

        def collect_lines(node: Node, indent: int):
            prefix = "    " * indent
            for l in node.lines:
                lines.append((prefix + l).rstrip())

            # remove trailing spacers
            children = node.children[:]
            while children and Tag.SPACER in children[-1].tags:
                children.pop()

            for child in children:
                if Tag.MODULE in node.tags:
                    nested_indent = indent
                else:
                    nested_indent = indent + 1
                collect_lines(child, nested_indent)

        collect_lines(out, 0)

        ret = "\n".join(lines) + "\n"
        return ret

    def stub(self, parent: Node, obj: ast.AST):
        match obj:
            # module entrypoint
            case ast.Module():
                self.stub_module(parent, obj)

            # imports
            case ast.Import() | ast.ImportFrom():
                self.stub_import(parent, obj)

            # assignments (eg constant definitions)
            case ast.Assign() | ast.AnnAssign():
                self.stub_assign(parent, obj)

            # class definition
            case ast.ClassDef():
                self.stub_class(parent, obj)

            # function definitions
            case ast.FunctionDef() | ast.AsyncFunctionDef():
                self.stub_func(parent, obj)

            # expressions (eg doc comments)
            case ast.Expr():
                self.stub_expr(parent, obj)

            # if
            case ast.If():
                self.stub_if(parent, obj)

            # try
            case ast.Try():
                self.stub_try(parent, obj)
            # pylint: disable-next=no-member
            case _ if sys.version_info >= (3, 11) and isinstance(obj, ast.TryStar):
                self.stub_try_star(parent, obj)

            # try
            case ast.With():
                self.stub_with(parent, obj)

            # type alias
            # pylint: disable-next=no-member
            case _ if sys.version_info >= (3, 12) and isinstance(obj, ast.TypeAlias):
                self.stub_type_alias(parent, obj)

            # intentionally ignored
            case (
                ast.Pass()
                | ast.For()
                | ast.Call()
                | ast.Assert()
                | ast.Delete()
                | ast.Raise()
                | ast.AugAssign()
                | ast.While()
                | ast.Match()
            ):
                self.log_stub_ignore(obj)

            # fallback
            case _:
                self.logger.ignore_unhandled(f"{type(obj).__name__} statement (`{self.unparse(obj)}`)")

    def stub_module(self, parent: Node, obj: ast.Module):
        body = obj.body
        type_ignores = obj.type_ignores
        if type_ignores:
            self.logger.ignore_unhandled("Type comment on module")

        this = parent.add_child(Tag.MODULE)
        for node in body:
            self.stub(this, node)

    @staticmethod
    def import_is_export(name: ast.alias) -> bool:
        if name.name == "*":
            return True

        if name.asname is None:
            return False

        return name.name == name.asname

    @staticmethod
    def get_imported_name(name: ast.alias) -> str:
        if name.asname is not None:
            return name.asname

        if "." not in name.name:
            return name.name

        return name.name.split(".")[0]

    def stub_import(self, parent: Node, obj: ast.Import | ast.ImportFrom):
        tags = {Tag.IMPORT}

        keep_unused_import = self.is_match(self.config.keep_unused_imports, parent, tags)
        if not keep_unused_import and self.used_names is not None:
            used_names = self.used_names

            new_names: list[ast.alias] = []
            for alias_name in obj.names:
                name = self.get_imported_name(alias_name)

                keep_import = name == "*" or name in used_names or self.import_is_export(alias_name)
                if keep_import:
                    new_names.append(alias_name)

            # everything got filtered out, drop the import statement
            if not new_names:
                return

            # modify import to remove the unused names
            obj.names.clear()
            obj.names.extend(new_names)

        line = self.unparse(obj)
        this = Node()
        this.add_tag(*tags)
        this.add_line(line)
        insert_pos = len(parent.children)

        parent.children.insert(insert_pos, this)

    # pylint: disable-next=too-many-locals
    def stub_assign(self, parent: Node, obj: ast.Assign | ast.AnnAssign):
        targets: list[ast.expr] | ast.Name | ast.Attribute | ast.Subscript
        value: ast.expr | None
        type_comment: str | None
        annotation: ast.expr | None
        simple: int | None
        tags = {Tag.VARIABLE}

        # normalize both cases into one
        match obj:
            case ast.Assign(targets, value, type_comment):
                annotation = None
                simple = None
            case ast.AnnAssign(target, annotation, value, simple):
                targets = target
                type_comment = None

        if type_comment:
            self.logger.ignore_unhandled("Type comment on variable")

        targets_unparsed = self.unparse_assign_target(targets, simple)
        if "." in targets_unparsed:
            # If we have an assignment of the form `x.y = ...`, then we interpret this
            # as a modification, not a definition, and
            # do not output this
            return

        # determine name of this variable.
        name = targets_unparsed.split("=")[0].strip()

        # node for this assignment
        this = Node(tags=tags)
        this.meta["name"] = name

        # annotation
        annotation_fragment = ""
        annotation_pattern = None
        if annotation is not None:
            annotation_unparsed = self.unparse_type_expr(annotation)
            annotation_fragment = ": " + annotation_unparsed
            this.add_tag(Tag.ANNOTATED)
            annotation_pattern = annotation_unparsed

        # value
        value_fragment = ""
        value_pattern = None
        if value is not None:
            value_unparsed = self.unparse(value)
            value_fragment = " = " + value_unparsed
            this.add_tag(Tag.ASSIGNED)
            value_pattern = value_unparsed

        keep_variable_value = self.is_match(
            self.config.keep_variable_value,
            parent,
            tags,
            name=name,
            annotation=annotation_pattern,
            value=value_pattern,
        )

        if not keep_variable_value and value_fragment != "":
            self.logger.ignore_disabled(f"Pruning value of variable {name}")
            value_fragment = " = ..."

        line = targets_unparsed + annotation_fragment + value_fragment
        this.add_line(line)

        keep_definitions = self.is_match(
            self.config.keep_definitions,
            parent,
            tags,
            name=name,
            annotation=annotation_pattern,
            value=value_pattern,
        )
        if not keep_definitions:
            self.logger.ignore_disabled(f"Pruning variable {name}")
            return

        parent.children.append(this)

    # pylint: disable-next=too-many-locals
    def stub_class(self, parent: Node, obj: ast.ClassDef):
        name = obj.name
        bases = obj.bases
        keywords = obj.keywords
        body = obj.body
        decorator_list = obj.decorator_list
        tags = {Tag.CLASS}

        if not self.is_match(self.config.keep_definitions, parent, tags, name=name):
            self.logger.ignore_disabled(f"Pruning class {name}")
            return

        this = parent.add_child(*tags)

        # decorators
        decorators_unparsed = [self.unparse(decorator) for decorator in decorator_list]
        for decorator in decorators_unparsed:
            this.add_line(f"@{decorator}")

        # generics
        generics = ""
        if sys.version_info >= (3, 12):
            generics = self.unparse_type_params(obj.type_params)

        # bases and keywords
        bases_unparsed = [self.unparse(base) for base in bases]
        for keyword in keywords:
            assert keyword.arg is not None
            bases_unparsed.append(f"{keyword.arg} = {self.unparse(keyword.value)}")

        sig_bases = ""
        if bases_unparsed:
            sig_bases = f"({', '.join(bases_unparsed)})"

        # signature
        this.add_line(f"class {name}{generics}{sig_bases}:")

        # recursively stub class contents
        for child_node in body:
            self.stub(this, child_node)

        # if we wrote no line, we need to add ...
        self.check_body_ellipsis(parent, this, name=name)

        # we add an empty line as a spacer after every function
        parent.add_child(Tag.SPACER).add_line("")

    # pylint: disable-next=too-many-locals
    def stub_func(self, parent: Node, obj: ast.FunctionDef | ast.AsyncFunctionDef):
        name: str = obj.name
        args: ast.arguments = obj.args
        body: list[ast.stmt] = obj.body
        decorator_list: list[ast.expr] = obj.decorator_list
        returns: ast.expr | None = obj.returns
        type_comment: str | None = obj.type_comment
        is_async_func: bool = isinstance(obj, ast.AsyncFunctionDef)
        tags = {Tag.FUNCTION}

        if not self.is_match(self.config.keep_definitions, parent, tags, name=name):
            self.logger.ignore_disabled(f"Pruning function {name}")
            return

        # commit to emitting a function
        this = parent.add_child(*tags)

        if type_comment:
            self.logger.ignore_unhandled("Type comment on function")

        # decorators
        decorators_unparsed = [self.unparse(decorator) for decorator in decorator_list]
        for decorator in decorators_unparsed:
            this.add_line(f"@{decorator}")

        # generics
        generics = ""
        if sys.version_info >= (3, 12):
            generics = self.unparse_type_params(obj.type_params)

        # signature
        args_unparsed = self.unparse_arguments(args)

        if returns is not None:
            returns_unparsed = " -> " + self.unparse_type_expr(returns)
        elif self.is_match(self.config.add_implicit_none_return, parent, tags, name=name):
            returns_unparsed = " -> None"
        else:
            returns_unparsed = ""

        signature = f"def {name}{generics}({args_unparsed}){returns_unparsed}:"
        if is_async_func:
            signature = "async " + signature
        this.add_line(signature)

        # check if we have a doc string in the body to keep
        if body and self.as_string_constant(body[0]) is not None:
            self.stub(this, body[0])

        # check if we have assignment statements, and inject them into
        # the surrounding class
        if (
            Tag.CLASS in parent.tags
            and name == "__init__"
            and self.is_match(self.config.add_class_attributes_from_init, parent, tags)
        ):
            self.check_add_class_attributes(parent, body)

        # make sure we have dots if there is nothing else
        self.check_body_ellipsis(parent, this, name=name)

        # we add an empty line as a spacer after every function
        parent.add_child(Tag.SPACER).add_line("")

    def stub_expr(self, parent: Node, obj: ast.Expr):
        match obj.value:
            case ast.Constant() if self.as_string_constant(obj.value) is not None:
                keep_docstring = False

                if not parent.children:
                    # regular docstring
                    keep_docstring = True
                else:
                    # potential inner docstring
                    preceding = parent.children[-1]
                    if self.is_match(self.config.keep_trailing_docstrings, parent, preceding.tags):
                        keep_docstring = True

                if keep_docstring:
                    raw_lit = self.source.unparse_original_source(obj)
                    parent.add_child(Tag.DOCSTRING).add_line(raw_lit, allow_multiline=True)
            case ast.Constant() if any(isinstance(obj.value.value, t) for t in CONSTANT_TYPES):
                self.log_stub_ignore(obj)
            case _:
                self.log_stub_ignore(obj)

    def stub_if(self, parent: Node, obj: ast.If):
        # we handle some ifs specially
        test = obj.test
        test_unparsed = self.unparse(test)

        tags = {Tag.IF}

        # lines guarded by "TYPE_CHECKING" are treated transparently
        skip_first_body = False
        if self.is_match(self.config.flatten_if, parent, tags, value=test_unparsed):
            # merge the if contents to the outside scope
            for child in obj.body:
                self.stub(parent, child)
            skip_first_body = True

        # we then emit the if itself with removed bodies
        if self.is_match(self.config.keep_if_statements, parent, tags=tags):
            this = parent.add_child(*tags)
            this.add_line(f"if {test_unparsed}:")

            if not skip_first_body:
                for child in obj.body:
                    self.stub(this, child)
            self.check_body_ellipsis(parent, this)

            orelse = obj.orelse

            # handle any intermediate elif cases
            while len(orelse) == 1 and isinstance(orelse[0], ast.If):
                test = orelse[0].test
                test_unparsed = self.unparse(test)

                this = parent.add_child(*tags)
                this.add_line(f"elif {test_unparsed}:")

                for child in orelse[0].body:
                    self.stub(this, child)
                self.check_body_ellipsis(parent, this)

                orelse = orelse[0].orelse

            # handle any final else
            if orelse:
                this = parent.add_child(*tags)
                this.add_line("else:")
                for child in orelse:
                    self.stub(this, child)
                self.check_body_ellipsis(parent, this)

    def stub_try(self, parent: Node, obj: ast.Try):
        # for try statements we just flatten the body into the parent
        for child in obj.body:
            self.stub(parent, child)

    if sys.version_info >= (3, 11):
        # pylint: disable-next=no-member
        def stub_try_star(self, parent: Node, obj: ast.TryStar):
            # for try statements we just flatten the body into the parent
            for child in obj.body:
                self.stub(parent, child)

    def stub_with(self, parent: Node, obj: ast.With):
        # for with statements we just flatten the body into the parent
        for child in obj.body:
            self.stub(parent, child)

    if sys.version_info >= (3, 12):
        # pylint: disable-next=no-member
        def stub_type_alias(self, parent: Node, obj: ast.TypeAlias):
            generics = self.unparse_type_params(obj.type_params)
            unparsed = f"type {obj.name.id}{generics} = {self.unparse_type_expr(obj.value)}"
            parent.add_child(Tag.TYPE_ALIAS).add_line(unparsed)

    def check_body_ellipsis(self, parent: Node, this: Node, name: str | None = None):
        if not this.children:
            this.add_child(Tag.ELLIPSIS).add_line("...")
        elif self.is_match(self.config.add_redundant_ellipsis, parent, this.tags, children=this.children, name=name):
            this.add_child(Tag.ELLIPSIS).add_line("...")

    def check_add_class_attributes(self, parent: Node, body: list[ast.stmt]):
        for body_stmt in body:
            if not isinstance(body_stmt, ast.AnnAssign):
                continue
            if not isinstance(body_stmt.target, ast.Attribute):
                continue

            target_value = body_stmt.target.value
            if not isinstance(target_value, ast.Name):
                continue

            if target_value.id != "self":
                continue

            attr = body_stmt.target.attr
            annot = self.unparse_type_expr(body_stmt.annotation)

            insert_pos = 0
            already_inserted = False
            for parent_child_idx, parent_child in enumerate(parent.children):
                if Tag.VARIABLE in parent_child.tags:
                    insert_pos = parent_child_idx + 1
                    if parent_child.meta["name"] == attr:
                        already_inserted = True
                elif Tag.DOCSTRING in parent_child.tags:
                    insert_pos = parent_child_idx + 1

            if not already_inserted:
                child_node = Node(tags={Tag.VARIABLE, Tag.ANNOTATED})
                child_node.add_line(f"{attr}: {annot}")
                child_node.meta["name"] = str(attr)
                parent.children.insert(insert_pos, child_node)

    def log_stub_ignore(self, obj: ast.AST):
        self.logger.ignore_intentional(f"{type(obj).__name__} statement")

    def as_string_constant(self, obj: ast.AST) -> str | None:
        if isinstance(obj, ast.Expr):
            obj = obj.value
        if isinstance(obj, ast.Constant) and isinstance(obj.value, str):
            return obj.value
        return None

    def unparse(self, obj: ast.AST) -> str:
        return self.source.unparse(obj)

    # pylint: disable-next=too-many-return-statements
    def unparse_type_expr(self, expr: ast.expr) -> str:
        """
        Convert an expression that is used in type annotation position
        to syntax appropriate for a typestub.

        Mostly this means identify string literals and replacing them with their content.
        """

        match expr:
            case ast.Subscript(value, slc, ast.Load()):
                prefix = self.unparse(value)
                if prefix == "Literal":
                    return self.source.unparse_original_source(expr)

                if isinstance(slc, ast.Tuple):
                    args = slc.elts
                else:
                    args = [slc]

                slc_unparsed_parts: list[str] = []
                for i, arg in enumerate(args):
                    if i > 0 and prefix == "Annotated":
                        slc_unparsed_parts.append(self.unparse(arg))
                    else:
                        slc_unparsed_parts.append(self.unparse_type_expr(arg))

                slc_unparsed = ", ".join(slc_unparsed_parts)
                return f"{prefix}[{slc_unparsed}]"
            case ast.Tuple(elts, ast.Load()):
                return "(" + ", ".join(self.unparse_type_expr(elt) for elt in elts) + ")"
            case ast.List(elts, ast.Load()):
                return "[" + ", ".join(self.unparse_type_expr(elt) for elt in elts) + "]"
            case ast.BinOp(left, ast.BitOr(), right):
                return self.unparse_type_expr(left) + " | " + self.unparse_type_expr(right)
            case ast.Constant(value) if isinstance(value, str):
                # check if we can parse the string as expression,
                # then return the unparsed expression
                try:
                    return self.unparse(self.source.parse_expr(value))
                except SyntaxError:
                    return self.unparse(expr)
            case _:
                # return unchanged
                return self.unparse(expr)

    def unparse_assign_target(
        self,
        targets: list[ast.expr] | ast.Name | ast.Attribute | ast.Subscript,
        simple: int | None,
    ) -> str:
        # target(s)
        if isinstance(targets, list):
            targets_unparsed = " = ".join([self.unparse(target) for target in targets])
        else:
            targets_unparsed = self.unparse(targets)
            if simple is not None and not simple:
                targets_unparsed = f"({targets_unparsed})"

        return targets_unparsed

    def unparse_arg(self, arg: ast.arg, val: None | ast.expr, keep_val: bool = False) -> str:
        if arg.type_comment:
            self.logger.ignore_unhandled("Type comment on function arg")

        ret = f"{arg.arg}"

        if arg.annotation is not None:
            ret += ": " + self.unparse_type_expr(arg.annotation)

        if val is not None:
            if keep_val:
                ret += " = " + self.unparse(val)
            else:
                ret += " = ..."

        return ret

    # pylint: disable-next=too-many-locals
    def unparse_arguments(self, ast_args: ast.arguments) -> str:
        posonlyargs = ast_args.posonlyargs
        args = ast_args.args
        vararg = ast_args.vararg
        kwonlyargs = ast_args.kwonlyargs
        kw_defaults = ast_args.kw_defaults
        kwarg = ast_args.kwarg
        defaults = ast_args.defaults

        def_counter = 0
        kw_def_counter = 0

        # the defaults variables contains the last N defaults, but for convenience
        # we compute an list where the initial skipped values are mapped to None
        aligned_defaults: list[None | ast.expr] = [None] * (len(posonlyargs) + len(args))
        gap = len(aligned_defaults) - len(defaults)
        aligned_defaults[gap:] = defaults

        unparsed_args: list[str] = []
        if posonlyargs:
            for arg in posonlyargs:
                val: ast.expr | None = None
                val = aligned_defaults[def_counter]
                def_counter += 1

                unparsed_args.append(self.unparse_arg(arg, val))
            unparsed_args.append("/")

        for arg in args:
            val: ast.expr | None = None
            val = aligned_defaults[def_counter]
            def_counter += 1

            unparsed_args.append(self.unparse_arg(arg, val))

        kwonlyargs_started = False

        if vararg is not None:
            unparsed_args.append("*" + self.unparse_arg(vararg, None))
            kwonlyargs_started = True

        if kwonlyargs:
            if not kwonlyargs_started:
                unparsed_args.append("*")
            for arg in kwonlyargs:
                val: ast.expr | None = None
                val = kw_defaults[kw_def_counter]
                kw_def_counter += 1

                unparsed_args.append(self.unparse_arg(arg, val))

        if kwarg is not None:
            unparsed_args.append("**" + self.unparse_arg(kwarg, None))

        ret = ", ".join(unparsed_args)
        return ret

    if sys.version_info >= (3, 12):
        # pylint: disable-next=no-member
        def unparse_type_params(self, type_params: ast.type_param) -> str:
            generics = ""
            if sys.version_info >= (3, 12):
                type_params_unparsed = []
                for type_param in type_params:
                    name = ""
                    bounds = ""
                    default = ""

                    match type_param:
                        # pylint: disable-next=no-member
                        case ast.TypeVar():
                            name = f"{type_param.name}"
                            if type_param.bound is not None:
                                bounds = f": {self.unparse_type_expr(type_param.bound)}"
                        # pylint: disable-next=no-member
                        case ast.TypeVarTuple():
                            name = f"*{type_param.name}"
                        # pylint: disable-next=no-member
                        case ast.ParamSpec():
                            name = f"**{type_param.name}"

                    if sys.version_info >= (3, 13) and type_param.default_value is not None:
                        default = f" = {self.unparse_type_expr(type_param.default_value)}"

                    type_params_unparsed.append(f"{name}{bounds}{default}")

                if type_params_unparsed:
                    generics = f"[{', '.join(type_params_unparsed)}]"
            return generics

    def _get_or_init_used_names(self) -> set[str]:
        if self.used_names is None:
            self.used_names = set()
        return self.used_names

    def discover_module_names(self, module: ast.Module):
        # discover all surviving names not brought into scope by imports
        for obj in module.body:
            match obj:
                case ast.Import() | ast.ImportFrom():
                    pass
                case ast.Assign() | ast.AnnAssign():
                    self.discover_any_names(obj)
                    self.discover_special_assign_names(obj)
                case _:
                    self.discover_any_names(obj)

    def discover_any_names(self, obj: ast.AST):
        names = self._get_or_init_used_names()

        for node in ast.walk(obj):
            if isinstance(node, ast.Name):
                names.add(node.id)

    def discover_special_assign_names(self, obj: ast.Assign | ast.AnnAssign):
        names = self._get_or_init_used_names()
        targets: list[ast.expr] | ast.Name | ast.Attribute | ast.Subscript
        value: ast.expr | None
        simple: int | None
        annotation: ast.expr | None

        # normalize both cases into one
        match obj:
            case ast.Assign(targets, value, _):
                simple = None
                annotation = None
            case ast.AnnAssign(target, annotation, value, simple):
                targets = target

        # check if we need to analyze the assignment value as a type
        if annotation is not None and value is not None:
            unparsed_annotation = self.unparse(annotation)
            if unparsed_annotation == "TypeAlias":
                re_parsed = self.source.parse_expr(self.unparse_type_expr(value))
                self.discover_any_names(re_parsed)

        # check if we need to analyze the assignment value as a list of types
        targets_unparsed = self.unparse_assign_target(targets, simple)
        if targets_unparsed == "__all__":
            match value:
                case ast.List(elts) | ast.Tuple(elts):
                    for elt in elts:
                        name = self.as_string_constant(elt)
                        if name is not None:
                            names.add(name)
                        else:
                            self.logger.ignore_unhandled(f"`__all__` list value {self.unparse(elt)} ({elt})")
                case _:
                    self.logger.ignore_unhandled(f"`__all__` value {self.unparse(value) if value else ''} ({value})")

    # pylint: disable-next=too-many-arguments
    def is_match(
        self,
        pattern: str | bool | None,
        parent: Node,
        tags: set[Tag],
        *,
        name: str | None = None,
        annotation: str | None = None,
        value: str | None = None,
        children: list[Node] | None = None,
    ) -> bool:

        child_tags: set[Tag] | None = None
        if children:
            child_tags = set()
            for child in children:
                child_tags.update(child.tags)

        ctx = MatchContext(
            parent_tags=parent.tags,
            tags=tags,
            file_path=str(self.source.relative_path),
            name=name,
            annotation=annotation,
            value=value,
            child_tags=child_tags,
        )
        return self.config.get_pattern(pattern).is_match(ctx)


def _stub_content(inp: str, relative_path: Path, config: ValidatedConfig, discover_imports: bool = False) -> str:
    source = Source(inp, relative_path, AstConfig(feature_version=config.get_python_version()))
    ast_module = source.parse_module()
    stubber = Stubber(source, config)
    if discover_imports:
        stubber.discover_module_names(ast_module)
    return stubber.output(ast_module)


def stubgen_single_file_src(inp: str, relative_path: Path, config: ValidatedConfig) -> str:
    # first generate the initial stub file
    stubbed = _stub_content(inp, relative_path, config)

    # then we re-stub the stub file, so that we can optionally analyze
    # imports and make sure we are idempotent
    stubbed = _stub_content(
        stubbed,
        relative_path,
        config,
        discover_imports=not config.get_pattern(config.keep_unused_imports).is_match(
            MatchContext(
                parent_tags={Tag.MODULE},
                tags={Tag.IMPORT},
                file_path=str(relative_path),
            )
        ),
    )

    return stubbed



from argparse import ArgumentParser

from ..config import main as config
from ..diff import main as diff
from ..evaluate import main as evaluate
from ..generate import main as generate
from ..util import extra_guard


def main():
    parser = ArgumentParser()

    subparsers = parser.add_subparsers(required=True)
    generate.register_args(subparsers.add_parser)
    evaluate.register_args(subparsers.add_parser)
    diff.register_args(subparsers.add_parser)
    config.register_args(subparsers.add_parser)

    args = parser.parse_args()

    with (
        extra_guard("def_fmt", executables=["black", "isort"]),
        extra_guard("eval", modules=["rich"], executables=["pyright", "stubgen"]),
    ):
        args.entrypoint(args)



import ast
from dataclasses import dataclass
from pathlib import Path
from typing import TypeAlias

SpannedNode: TypeAlias = ast.stmt | ast.expr


@dataclass
class AstConfig:
    feature_version: tuple[int, int] | None = None
    type_comments: bool = False


class Source:
    src: bytes
    src_decoded: str
    relative_path: Path
    filename: str
    line_offsets: list[tuple[int, int]]
    config: AstConfig

    def __init__(self, src: str, relative_path: Path, config: AstConfig):
        self.src = src.encode()
        self.src_decoded = src
        self.relative_path = relative_path
        self.filename = relative_path.name
        self.line_offsets = []
        self.config = config

        offset = 0
        for line in self.src.splitlines(keepends=True):
            end = offset + len(line)
            self.line_offsets.append((offset, end))
            offset = end

        assert offset == len(self.src)

    def parse_module(self) -> ast.Module:
        module: ast.Module = ast.parse(
            self.src_decoded,
            self.filename,
            feature_version=self.config.feature_version,
            type_comments=self.config.type_comments,
        )
        assert isinstance(module, ast.Module)
        return module

    def parse_standalone_expr(self, src: str) -> ast.Expression:
        expr: ast.Expression = ast.parse(
            src,
            mode="eval",
            feature_version=self.config.feature_version,
            type_comments=self.config.type_comments,
        )
        assert isinstance(expr, ast.Expression)
        return expr

    def parse_expr(self, src: str) -> ast.expr:
        return self.parse_standalone_expr(src).body

    def unparse_offsets(
        self, node: SpannedNode, include_decorators: bool = True, verify: bool = False
    ) -> tuple[int, int]:
        # check invariants
        assert node.lineno is not None
        assert node.end_lineno is not None
        assert node.col_offset is not None
        assert node.end_col_offset is not None

        # process position data
        lineno = node.lineno - 1
        end_lineno = node.end_lineno - 1
        col_offset = node.col_offset
        end_col_offset = node.end_col_offset

        # decorators are not included in the span of a class or function,
        # so we need to manually extend the span here
        if include_decorators:
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
                for decorator in node.decorator_list:
                    decorator_lineno = decorator.lineno - 1
                    decorator_col_offset = decorator.col_offset

                    # check if we got a decorator on an earlier syntactic position
                    if decorator_lineno < lineno:
                        lineno = decorator_lineno
                        # NB: The prefix "@" is not covered by the expression,
                        # so we need to manually step back
                        col_offset = decorator_col_offset - 1
                        assert col_offset >= 0

        # start of syntax element
        start_offset = self.line_offsets[lineno][0] + col_offset

        # end of syntax element
        end_offset = self.line_offsets[end_lineno][0] + end_col_offset

        # check against python impl
        if verify:
            sliced = self.src[start_offset:end_offset].decode()
            unparsed = ast.get_source_segment(self.src_decoded, node)
            assert unparsed == sliced, "\n".join(
                [
                    "",
                    f"sliced:   >{sliced}<",
                    f"unparsed: >{unparsed}<",
                ]
            )

        # return start and end
        return (start_offset, end_offset)

    def unparse(self, node: ast.AST) -> str:
        return ast.unparse(node)

    def unparse_original_source(self, node: SpannedNode, verify: bool = False) -> str:
        start, end = self.unparse_offsets(node, verify=verify)
        return self.src[start:end].decode()



import json
from textwrap import dedent

from .. import toml
from ..util import Json
from . import Config
from .pattern import Pattern


def fmt_config_dict(config: Config, show_format: str) -> dict[str, Json]:
    re_parse = show_format != "toml"

    out: dict[str, Json] = {}
    for field_name, field in Config.get_fields().items():
        val = config.get(field_name)
        if val is None:
            continue

        if field.validated_ty.ty is Pattern:
            raw = val
            if isinstance(raw, bool):
                pass
            elif raw == "True":
                raw = True
            elif raw == "False":
                raw = False
            else:
                assert isinstance(raw, str)
                if not re_parse:
                    raw = dedent(raw).strip()
                    lines = raw.splitlines()
                    raw = "".join(f"    {line}\n" for line in lines)
        elif isinstance(val, list):
            assert field.validated_ty.generic_string() == "list[FormatterCmd]"
            raw = [
                {
                    "name": command.name,
                    "cmdline": command.cmdline,
                }
                for command in val
            ]
        else:
            raw = val

        out[field_name] = raw

    return out


def fmt_config(config: Config, show_format: str) -> str:
    out = fmt_config_dict(config, show_format)

    ret = ""
    if show_format == "plain":
        for field, raw in out.items():
            ret += f"{field}={raw}\n"
    elif show_format == "json":
        out_with_header = {"tool": {"dubstub": out}}
        ret = json.dumps(out_with_header, indent=4)
    elif show_format == "toml":
        out_with_header = {"tool": {"dubstub": out}}
        ret = toml.dumps(out_with_header, indent=4, multiline_strings=True)

    return ret.strip()



import sys
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from types import NoneType, UnionType
from typing import Annotated, Any, TypeAlias, get_args, get_origin, get_type_hints

from .. import toml
from .pattern import Pattern


@dataclass
class FormatterCmd:
    name: str
    cmdline: list[str]


RawConfigType: TypeAlias = str | bool | list[FormatterCmd]
ValidatedConfigType: TypeAlias = str | Pattern | list[FormatterCmd]


@dataclass
class TypeModel:
    ty: type
    args: list["TypeModel"]

    @staticmethod
    def parse_annotation(annotation: Any) -> "TypeModel":
        origin = get_origin(annotation)
        if origin is None:
            return TypeModel(annotation, [])

        args = get_args(annotation)
        return TypeModel(origin, [TypeModel.parse_annotation(arg) for arg in args])

    def generic_string(self) -> str:
        ret = self.ty.__name__
        if self.args:
            ret += f"[{', '.join(arg.generic_string() for arg in self.args)}]"
        return ret

    def args_of(self, ty: Any) -> list["TypeModel"]:
        assert self.ty is ty
        return self.args


@dataclass
class Field:
    raw_tys: list[TypeModel]
    validated_ty: TypeModel
    description: str

    @staticmethod
    def parse_annotation(annotation: Any) -> "Field":
        model = TypeModel.parse_annotation(annotation)

        # assert that we got an Annotation and get their args
        raw_union_ty, validated_ty, description_ = model.args_of(Annotated)

        # get description and reformat it
        description = description_.ty
        assert isinstance(description, str)
        description = dedent(description).strip()

        # assert that we got an Union and get their args
        raw_tys = [raw_ty for raw_ty in raw_union_ty.args_of(UnionType) if raw_ty.ty is not NoneType]
        return Field(raw_tys, validated_ty, description)


@dataclass
class Config:
    # pylint: disable=too-many-instance-attributes

    profile: Annotated[
        str | None,
        str,
        """
        Base profile to use for config settings.

        Any explicitly set config values will overwrite those defined by the profile.
        """,
    ] = None

    python_version: Annotated[
        str | None,
        str,
        """
        Version of python syntax to target.

        Valid forms are `"X.Y"` and `"auto"`, the latter of which derives
        the version from the current python interpreter.
        """,
    ] = None

    keep_definitions: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to keep a class, function, or variable definition during type stubbing.

        This can be used to remove private elements (ie, those starting with a single `_`).
        """,
    ] = None

    keep_trailing_docstrings: Annotated[
        str | bool | None,
        Pattern,
        '''
        Whether to keep a trailing docstring for an element.

        A trailing docstring is defined as a string literal that is not the first
        statement in a module, class or function block.

        These kinds of docstrings are supported by some documentation generators and IDE
        plugins.

        Example:
        ```python
        class foo:
            """normal docstring"""
            x: y = z
            """trailing docstring"""
        ```
        ''',
    ] = None

    add_implicit_none_return: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to add a `-> None` return type to a function that has not specified a
        return type annotation.
        """,
    ] = None

    keep_if_statements: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to keep an `if` statement.

        If an `if` statement is kept, its condition expression will be kept fully,
        and its body will be recursively stubbed.
        """,
    ] = None

    flatten_if: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to flatten the first body of an `if` into the surrounding scope.

        This setting combines with `keep_if_statements`: If an `if` statement is matched
        by both, the first body of the kept `if` will be empty.

        This does not apply to `elif` cases of an `if`.
        """,
    ] = None

    add_redundant_ellipsis: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to add redundant ellipsis (`...`) statements to bodies of statements that are not empty after stubbing.

        Example:
        ```python
        class foo:
            ... # always added
        class bar:
            x = y
            ... # redundant
        ```
        """,
    ] = None

    keep_variable_value: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to keep the assignment value of a variable definition.
        If the value is not kept, it is replaced with an ellipsis (`...`).

        Example:
        ```python
        foo = 42        # value kept
        foo: bar = 42   # value kept
        foo = ...       # value not kept
        foo: bar = ...  # value not kept
        ```
        """,
    ] = None

    keep_unused_imports: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to keep import statements that seem to not be used anywhere
        in the module.

        Note that this is a heuristic based on searching for the imported name,
        it can have false positives, which erroneously consider an import used.

        Both normal python code, as well the contents of `__all__` list strings will be considered.

        Certain imports will be interpreted as re-exports, and always considered used.

        For example, given this code:
        ```python
        x: Foo
        ```
        Then imports would be considered like this:
        ```python
        from foo import Foo   # considered used
        from bar import Bar   # considered not used
        import X as X         # considered used (always)
        import a.b.X as X     # considered not used
        from Y import X as X  # considered used (always)
        from Y import *       # considered used (always)
        ```

        See also https://typing.readthedocs.io/en/latest/spec/distributing.html#import-conventions
        """,
    ] = None

    add_class_attributes_from_init: Annotated[
        str | bool | None,
        Pattern,
        """
        Whether to class attribute type annotations from `self.<name>: T` annotations
        in `__init__()` methods.
        """,
    ] = None

    format: Annotated[
        str | bool | None,
        Pattern,
        """
        Enable autoformatting of the generated typestub code.

        The default output of this tool follows no consistent formatting style:
        Individual lines either contain a single large statement, or are copy-pasted from the original source verbatim.

        If enabled, the generated or copied `.pyi` files will be passed to the commands specified by `formatter_cmds`.
        """,
    ] = None

    formatter_cmds: Annotated[
        list[FormatterCmd] | None,
        list[FormatterCmd],
        """
        List of formatter cmds that will be executed on generated and copied `.pyi` files.

        Each `FormatterCmd` is an object with an arbitrary `name`, and a `cmdline` list
        that defines how to execute the command:

        ```toml
        [[tool.dubstub.formatter_cmds]]
        name = "<cmd name>"
        cmdline = [
            "<cmd>",
            "<arg1>",
            "<arg2>",
            ...
        ]
        ```

        There are a number of variable substitutions supported for the `cmdline` arguments:

        - `${dubstub_py_major}` will be replaced by the major version of python selected
           with the `python_version` config setting.
        - `${dubstub_py_minor}` will be replaced by the minor version of python selected
           with the `python_version` config setting.
        - `${dubstub_py_exe}` will be replaced by the path to the current python executable (`sys.executable`).
        - `${dubstub_file_arg}` will be replaced with the path to a file that should be formatted in-place.
        - `${dubstub_file_args}` will be replaced with the path to a file that should be formatted in-place,
          and will also cause the argument to expand into multiple arguments, one for each file.

        The commands will be executed in series, with the output of one feeding into the next one.
        The difference between the last two substitution options is that `dubstub_file_arg` will call the command
        once for every single file, while `dubstub_file_args` will batch multiple files into a single command call.

        Examples:

        - `cmdline = ["black", "${dubstub_file_arg}"]`
          will cause these command calls:
          ```bash
          black file1.pyi
          black file2.pyi
          black file3.pyi
          ...
          ```
        - `cmdline = ["black", "${dubstub_file_args}"]`
          will cause these command calls:
          ```bash
          black file1.pyi file2.pyi file3.pyi ...
          ```
        """,
    ] = None

    @staticmethod
    def get_fields() -> dict[str, Field]:
        """
        Returns a mapping of field name to field type and description text.
        """

        ret: dict[str, Any] = {}

        for field_name, annotation in get_type_hints(Config, include_extras=True).items():
            if get_origin(annotation) is Annotated:
                ret[field_name] = Field.parse_annotation(annotation)

        return ret

    @staticmethod
    def _check_name(name: str):
        if name not in Config.get_fields():
            raise ValueError(f"unknown config setting {name}")

    def get(self, name: str) -> RawConfigType | None:
        """Get the value of the field with the give name"""

        self._check_name(name)
        return getattr(self, name)

    def set(self, name: str, value: RawConfigType | None):
        """Set the value of the field with the give name"""

        self._check_name(name)
        fields = self.get_fields()

        is_raw_ty = any(isinstance(value, raw_ty.ty) for raw_ty in fields[name].raw_tys)

        if not (value is None or is_raw_ty):
            raise ValueError(f"wrong type for config setting {name} (got {type(value).__name__})")

        setattr(self, name, value)

    @staticmethod
    def parse_config(
        path: Path | None = None,
        raw: str | None = None,
        obj: dict[str, Any] | None = None,
    ) -> "Config":
        """
        Parse a config from one of these sources:
        - a path to a `.toml` file with `tool.dubstub` keys.
        - a raw toml string representation with `tool.dubstub` keys.
        - a dictionary with `tool.dubstub` keys.
        """

        config = Config()

        if obj is None:
            if raw is None:
                if path is None:
                    raise ValueError("one of the function arguments needs to be set")
                raw = path.read_text()
            obj = toml.loads(raw)

        raw_config = obj.get("tool", {}).get("dubstub", {})
        for field, val in raw_config.items():
            if field not in Config.get_fields():
                raise ValueError(f"unknown config key {field}")
            config.set(field, val)

        return config

    def validate(self) -> "ValidatedConfig":
        """
        Validate the Config, which involves additional checks,
        and filling in all missing data with default values from the selected
        profile.
        """

        validated = ValidatedConfig()
        validated.profile = self.profile or "default"

        for field, meta in self.get_fields().items():
            if field == "profile":
                continue

            value = self.get(field)
            if value is None:
                value = PROFILES[validated.profile].fields.get(field)
            if value is None:
                value = PROFILES["default"].fields.get(field)
            assert value is not None, "value not found in default profile"

            validated.set(field, value)

            if meta.validated_ty.ty is Pattern:
                assert isinstance(value, (bool, str))
                validated.pattern[value] = Pattern(value)

        # check that we did not forget about any field
        for field in validated.get_fields():
            assert validated.get(field) is not None

        # check that the python version has a valid format
        validated.get_python_version()

        return validated


class ValidatedConfig(Config):
    pattern: dict[str | bool, Pattern]

    def __init__(self):
        super().__init__()
        self.pattern = {}

    def get_pattern(self, raw: str | bool | None) -> Pattern:
        """get a compiled pattern by its raw value"""

        assert raw is not None
        return self.pattern[raw]

    def get_python_version(self) -> tuple[int, int]:
        """return the parsed python version"""

        assert self.python_version is not None

        if self.python_version == "auto":
            return sys.version_info[0:2]

        parts = self.python_version.split(".")
        assert len(parts) == 2

        return (int(parts[0]), int(parts[1]))

    def get_formatting(self) -> tuple[Pattern, list[FormatterCmd]]:
        """return formatting configs"""

        assert self.formatter_cmds is not None
        return self.get_pattern(self.format), self.formatter_cmds


@dataclass
class Profile:
    fields: Config
    description: str


PROFILES: dict[str, Profile] = {
    "default": Profile(
        Config(
            python_version="auto",
            keep_definitions=r"""
                not name_is('_[^_].*')
            """,
            keep_trailing_docstrings=r"""
                node_is('variable')
            """,
            add_implicit_none_return=r"""
                parent_node_is('class') and name_is('__init__')
            """,
            keep_if_statements=True,
            flatten_if=r"""
                value_is('TYPE_CHECKING')
            """,
            add_redundant_ellipsis=False,
            keep_variable_value=r"""
                annotation_is('TypeAlias|([tT]ype(\\[.*\\])?)')
                or value_is('(TypeVar|TypeVarTuple|ParamSpec)\\(.*\\)')
                or value_is('(NamedTuple|NewType|TypedDict)\\(.*\\)')
                or (parent_node_is('module') and name_is('__all__'))
                or (parent_node_is('class') and name_is('__model__'))
            """,
            keep_unused_imports=False,
            add_class_attributes_from_init=True,
            format=False,
            formatter_cmds=[
                FormatterCmd(
                    name="isort",
                    cmdline=[
                        # "${dubstub_py_exe}",
                        # "-m",
                        "isort",
                        "--py",
                        "${dubstub_py_major}${dubstub_py_minor}",
                        "--profile",
                        "black",
                        "--settings",
                        "/dev/null",
                        "${dubstub_file_args}",
                    ],
                ),
                FormatterCmd(
                    name="black",
                    cmdline=[
                        # "${dubstub_py_exe}",
                        # "-m",
                        "black",
                        "--pyi",
                        "--target-version",
                        "py${dubstub_py_major}${dubstub_py_minor}",
                        "--skip-magic-trailing-comma",
                        "--fast",
                        "--config",
                        "/dev/null",
                        "${dubstub_file_args}",
                    ],
                ),
            ],
        ),
        """
        The default profile tries to have sensible defaults that match
        the official recommendations:

        - Private definitions are removed, and identified by exactly one leading `_`.
        - Trailing docstrings are kept for variables.
        - We add a `-> None` return type on functions that should have them.
        - `if` statements are kept, but `TYPE_CHECKING` guards are merged into the surrounding scope.
        - Variable values are kept if they define types according to the `typing` module.
        - Class attributes are also looked for in `__init__()` method assignments.
        - Unused imports are remove.
        - Autoformatting is disabled, but will use isort and black with default settings if enabled.
        """,
    ),
    "no_privacy": Profile(
        Config(
            keep_definitions=True,
            keep_unused_imports=True,
        ),
        """
        This profile is the same as `default`, but with privacy disabled:

        - No definitions with "private" names are filtered out.
        - No unused imports are filtered out.
        """,
    ),
    "pyright": Profile(
        Config(
            keep_definitions=r"""
                not (node_is('function') and name_is('_[^_].*'))
            """,
            add_redundant_ellipsis=r"""
                not any_child_node_is('function|assigned|class')
            """,
            keep_trailing_docstrings=False,
        ),
        """
        This profiles tries to approximate `pyright`s stubgen behavior better:

        - Only private functions are removed.
        - Redundant ellipsis are added to a few locations.
        - Trailing docstrings are removed.
        """,
    ),
}



import shlex
import subprocess
import sys
from pathlib import Path

from .config import FormatterCmd, ValidatedConfig
from .config.match_ctx import MatchContext, Tag
from .fs import Walker


def _generate_jobs(command: FormatterCmd, py_major: int, py_minor: int, paths: set[Path]) -> list[FormatterCmd]:
    jobs: list[FormatterCmd] = []

    has_dubstub_file_arg = any("${dubstub_file_arg}" in arg for arg in command.cmdline)
    has_dubstub_file_args = any("${dubstub_file_args}" in arg for arg in command.cmdline)
    assert (
        has_dubstub_file_arg != has_dubstub_file_args
    ), "exactly one of `${dubstub_file_arg}` or `${dubstub_file_args}` needs to be used in the cmdline"

    def common_replace(arg: str) -> str:
        arg = arg.replace("${dubstub_py_major}", str(py_major))
        arg = arg.replace("${dubstub_py_minor}", str(py_minor))
        arg = arg.replace("${dubstub_py_exe}", sys.executable)
        return arg

    if has_dubstub_file_arg:
        for path in sorted(paths):
            cmd: list[str] = []
            for arg in command.cmdline:
                arg = common_replace(arg)
                arg = arg.replace("${dubstub_file_arg}", str(path))
                cmd.append(arg)
            jobs.append(FormatterCmd(name=command.name, cmdline=cmd))
    else:
        cmd: list[str] = []
        for arg in command.cmdline:
            arg = common_replace(arg)
            if "${dubstub_file_args}" in arg:
                for path in sorted(paths):
                    arg_instance = arg.replace("${dubstub_file_args}", str(path))
                    cmd.append(arg_instance)
            else:
                cmd.append(arg)
        jobs.append(FormatterCmd(name=command.name, cmdline=cmd))

    return jobs


def format_pyi_tree(walker: Walker, config: ValidatedConfig):
    format_, formatter_cmds = config.get_formatting()
    py_major, py_minor = config.get_python_version()

    # discover the paths of all pyi files we need to process
    paths: set[Path] = set()

    for event in walker.walk():
        path = event.out_path

        if path in paths:
            continue

        if path.suffix != ".pyi":
            continue

        if not path.is_file():
            continue

        ctx = MatchContext(
            set(),
            {Tag.MODULE},
            event.out_rel_pattern,
        )

        if not format_.is_match(ctx):
            continue

        paths.add(path)

    if not paths:
        return

    print(f"Run formatter commands on {len(paths)} files")
    failed = False

    jobs: list[FormatterCmd] = []
    for command in formatter_cmds:
        jobs.extend(_generate_jobs(command, py_major, py_minor, paths))

    for command in jobs:
        result = subprocess.run(
            command.cmdline,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
            check=False,
        )

        if result.returncode != 0:
            print(f"{command.name} command failed:")
            print(shlex.join(command.cmdline))
            print("- output ---")
            print(result.stdout)
            print("------------")
            failed = True
            break

    assert not failed, "formatter failed"



import ast
from pathlib import Path
from types import CodeType
from typing import Any, Callable, TypeGuard

from ..source import AstConfig, Source
from ..util import regex_match
from .match_ctx import MatchContext

FUNCTIONS = [
    "parent_node_is",
    "node_is",
    "file_path_is",
    "name_is",
    "annotation_is",
    "value_is",
    "any_child_node_is",
]


class AstValidator:
    reasons: list[str]

    def __init__(self):
        self.reasons = []

    def fail(self, reason: str) -> bool:
        self.reasons = [reason]
        return False

    def is_allowed(self, node: ast.AST) -> bool:
        match node:
            case ast.Constant() if self.is_simple_constant(node, bool):
                return True
            case ast.Call() if self.is_allowed_call(node):
                return True
            case ast.BoolOp(ast.And() | ast.Or(), values) if all(self.is_allowed(value) for value in values):
                return True
            case ast.UnaryOp(ast.Not(), value) if self.is_allowed(value):
                return True
            case _:
                self.reasons.append(f"unsupported type: ast.{type(node).__name__}")
                return False

    def is_simple_constant(self, node: ast.AST, value_type: Any) -> bool:
        if not isinstance(node, ast.Constant):
            return self.fail(f"expected ast.Constant, got: {type(node).__name__}")

        if not isinstance(node.value, value_type):
            return self.fail(f"unsupported constant type: {type(node.value).__name__}")

        if node.kind is not None:
            return self.fail(f"unsupported node kind: {node.kind}")

        return True

    def is_allowed_name(self, node: ast.AST) -> TypeGuard[ast.Name]:
        if not isinstance(node, ast.Name):
            return self.fail("expected ast.Name")

        if not isinstance(node.ctx, ast.Load):
            return self.fail("expected ast.Name with ctx ast.Load")

        return True

    def is_allowed_call(self, node: ast.Call) -> bool:
        if not self.is_allowed_name(node.func):
            return self.fail("unsupported function syntax")

        if node.func.id not in FUNCTIONS:
            return self.fail("unsupported function name")

        if node.keywords or len(node.args) != 1:
            return self.fail("unsupported function signature")

        arg0 = node.args[0]
        if not self.is_simple_constant(arg0, str):
            return self.fail("unsupported function argument")
        return True


def make_pattern_eval(compiled: CodeType) -> Callable[[MatchContext], bool]:
    def inner(ctx: MatchContext) -> bool:
        class Funcs:
            @staticmethod
            def parent_node_is(pat: str) -> bool:
                return regex_match(pat, [tag.value for tag in ctx.parent_tags])

            @staticmethod
            def node_is(pat: str) -> bool:
                return regex_match(pat, [tag.value for tag in ctx.tags])

            @staticmethod
            def name_is(pat: str) -> bool:
                return ctx.name is not None and regex_match(pat, ctx.name)

            @staticmethod
            def file_path_is(pat: str) -> bool:
                return regex_match(pat, ctx.file_path)

            @staticmethod
            def annotation_is(pat: str) -> bool:
                return ctx.annotation is not None and regex_match(pat, ctx.annotation)

            @staticmethod
            def value_is(pat: str) -> bool:
                return ctx.value is not None and regex_match(pat, ctx.value)

            @staticmethod
            def any_child_node_is(pat: str) -> bool:
                return ctx.child_tags is not None and regex_match(pat, [tag.value for tag in ctx.child_tags])

        env = {func: getattr(Funcs, func) for func in FUNCTIONS}

        # pylint: disable-next=eval-used
        ret = eval(compiled, env, env)
        assert isinstance(ret, bool)
        return ret

    return inner


def parse_pattern(pattern: bool | str) -> tuple[Callable[[MatchContext], bool], ast.Expression]:
    source = Source("", Path(), AstConfig())

    if isinstance(pattern, bool):
        pattern = repr(pattern)

    parsed = source.parse_standalone_expr(f"({pattern})")
    validator = AstValidator()
    if not validator.is_allowed(parsed.body):
        raise ValueError(
            f"Pattern is not allowed to contain Python AST element: `{source.unparse(parsed)}` ({', '.join(validator.reasons)})"
        )
    compiled = compile(parsed, "<pattern>", mode="eval")
    return make_pattern_eval(compiled), parsed


class Pattern:
    pattern: Callable[[MatchContext], bool]
    _raw: bool | str
    _parsed: ast.Expression

    def __init__(self, pattern: bool | str):
        self.pattern, self._parsed = parse_pattern(pattern)
        self._raw = pattern

    def is_match(self, ctx: MatchContext) -> bool:
        return self.pattern(ctx)

    def __bool__(self):
        raise TypeError

    def get_raw(self, reformat: bool = False) -> str:
        if reformat:
            return ast.unparse(self._parsed)
        return str(self._raw)

    def __str__(self) -> str:
        return ast.unparse(self._parsed)

    def __repr__(self) -> str:
        return f"Pattern({repr(str(self))})"



from argparse import SUPPRESS, ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path
from typing import Callable

from ..cli.common import add_common_options, parse_config_from_cli


def register_args(make_parser: Callable[..., ArgumentParser]):
    parser = make_parser("config", formatter_class=RawTextHelpFormatter)
    parser.set_defaults(entrypoint=main)
    parser.add_argument("--show", action="store_true", help="show effective config")
    parser.add_argument("--show-format", default="plain", choices=["plain", "json", "toml"])
    parser.add_argument("--gen-docs", type=Path, help=SUPPRESS)
    add_common_options(parser)


def main(args: Namespace):
    config = parse_config_from_cli(args)

    if args.gen_docs:
        from .gen_docs import generate_docs  # pylint:disable=import-outside-toplevel

        generate_docs(args.gen_docs)
    elif args.show or args.show_format:
        from .show import fmt_config  # pylint:disable=import-outside-toplevel

        validated = config.validate()
        print(fmt_config(validated, args.show_format))



from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path
from typing import Callable

from ..cli.common import add_common_options, parse_config_from_cli
from ..util import extra_guard
from .fs import generate_stubs


def register_args(make_parser: Callable[..., ArgumentParser]):
    parser = make_parser("gen", formatter_class=RawTextHelpFormatter)
    parser.set_defaults(entrypoint=main)
    parser.add_argument("--input", type=Path, required=True, help="Input file or directory tree")
    parser.add_argument("--output", type=Path, required=True, help="Output file or directory tree")
    add_common_options(parser)


def main(args: Namespace):
    with extra_guard("def_fmt", executables=["black", "isort"]):
        generate_stubs(
            args.input,
            args.output,
            config=parse_config_from_cli(args),
        )



from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path
from types import ModuleType
from typing import Callable

from ..cli.common import add_common_options, parse_config_from_cli
from ..config import Config
from . import gen_dubstub, gen_mypy, gen_pyright

GENERATORS: dict[str, ModuleType] = {
    "dubstub": gen_dubstub,
    "pyright": gen_pyright,
    "mypy": gen_mypy,
}


def register_args(make_parser: Callable[..., ArgumentParser]):
    parser = make_parser("eval", formatter_class=RawTextHelpFormatter)
    parser.set_defaults(entrypoint=main)
    parser.add_argument("--input", type=Path, required=True, help="Input file or directory tree")
    parser.add_argument("--output", type=Path, required=True, help="Output file or directory tree")
    parser.add_argument(
        "--type",
        choices=["dubstub", "pyright", "mypy", "all"],
        default="all",
        help="Which typestub generators to run. Their output will be placed side-by-side in the output directory.",
    )
    parser.add_argument("--group", help="Optional extra directory name to group output of multiple calls under.")
    add_common_options(parser)


def main(args: Namespace):
    run(
        args.input,
        args.output,
        group=args.group,
        generators=[args.type],
        config=parse_config_from_cli(args),
    )


# pylint: disable-next=too-many-arguments,too-many-locals
def run(
    input_path: Path,
    output_path: Path,
    *,
    group: str | None = None,
    generators: list[str] | None = None,
    config: Config | None = None,
):
    # NB: We want to support merging namespace module directories, so
    # we actually descent down into all non-namespace modules when processing the input directory
    inp: Path = input_path.resolve()
    out: Path = output_path.resolve()

    validated = (config or Config()).validate()

    def get_out_sub(generator_name: str):
        out_sub = out / generator_name
        if group:
            out_sub = out_sub / group
        return out_sub

    # next to dubstub itself, we support other typestub generators for comparison
    selected_generators: dict[str, ModuleType] = {}
    for generator in generators or []:
        if generator == "all":
            for generator_name, generator_module in GENERATORS.items():
                selected_generators[generator_name] = generator_module
        else:
            selected_generators[generator] = GENERATORS[generator]

    # For each generator, generate type stubs for each discovered input child directory
    for generator_name, generator_module in selected_generators.items():
        print(f"Generate typestubs with {generator_name}")

        # determine the exact output directory for this generator
        out_sub = get_out_sub(generator_name)

        # actually generate the type stubs
        generator_module.generate(inp, out_sub, validated)
        print()



from pathlib import Path

from ..config import ValidatedConfig
from ..generate.fs import generate_stubs


def generate(inp: Path, out: Path, config: ValidatedConfig):
    generate_stubs(inp, out, config)



from pathlib import Path
from textwrap import dedent

from pycmarkgfm import gfm_to_html, options  # type: ignore

from . import PROFILES, Config
from .show import fmt_config


def render(gfm: str) -> str:
    return gfm_to_html(dedent(gfm), options.github_pre_lang)  # type: ignore


class MarkdownTableRender:
    cols: list[str]
    rows: list[list[str]]

    def __init__(self, *cols: str):
        self.cols = list(cols)
        self.rows = []

    def add_row(self, *cells: str):
        self.rows.append(list(cells))

    def render(self) -> str:
        lookup: dict[str, str] = {}

        # first render the table to html
        ret = ""
        ret += f"| { ' | ' .join(self.cols) } |\n"
        ret += "|" + (" --- |" * len(self.cols)) + "\n"

        for row in self.rows:
            ret += "|"
            for cell in row:
                key = f"CELL[{len(lookup)}]"
                rendered = render(cell)
                lookup[key] = rendered
                ret += f" {key} |"
            ret += "\n"

        ret = render(ret)

        # then insert the rendered cells into it
        for key, val in lookup.items():
            ret = ret.replace(key, val)

        return ret.rstrip()


def generate_config_docs(content: str) -> str:
    table = MarkdownTableRender("Config Setting", "Description")
    for field_name, field in Config.get_fields().items():
        table.add_row(
            f"`{field_name}`\n\ntype: `{field.validated_ty.generic_string()}`",
            field.description,
        )

    start_marker = "<!-- CONFIG_START -->\n"
    start_pos = content.find(start_marker)
    assert start_pos >= 0
    start_pos += len(start_marker)

    end_marker = "\n<!-- CONFIG_END -->"
    end_pos = content.find(end_marker)
    assert end_pos >= 0

    rendered = table.render()
    return content[:start_pos] + rendered + content[end_pos:]


def generate_profile_docs(content: str) -> str:
    rendered = ""
    for profile_name, profile_data in PROFILES.items():
        description = dedent(profile_data.description).strip()

        rendered += f"#### Profile `{profile_name}`\n"
        rendered += "\n"
        rendered += description.rstrip() + "\n"
        rendered += "\n"
        rendered += "Exact config settings set by the profile:\n"
        rendered += "\n"
        rendered += "```toml\n" + fmt_config(profile_data.fields, "toml") + "\n```\n"

    start_marker = "<!-- PROFILE_START -->\n"
    start_pos = content.find(start_marker)
    assert start_pos >= 0
    start_pos += len(start_marker)

    end_marker = "\n<!-- PROFILE_END -->"
    end_pos = content.find(end_marker)
    assert end_pos >= 0

    return content[:start_pos] + rendered + content[end_pos:]


def generate_docs(out: Path):
    if not out.exists():
        out.write_text(
            dedent(
                """
                    # debug config
                    <!-- CONFIG_START -->
                    <!-- CONFIG_END -->

                    # debug profiles
                    <!-- PROFILE_START -->
                    <!-- PROFILE_END -->
                """
            )
        )

    content = out.read_text()
    content = generate_config_docs(content)
    content = generate_profile_docs(content)
    out.write_text(content)



from argparse import ArgumentParser, Namespace
from pathlib import Path

from ..config import Config
from ..config.pattern import Pattern


def add_common_options(parser: ArgumentParser):
    parser.add_argument("--config", type=Path, help="Path to a config file")
    group = parser.add_argument_group("config options")
    for field_name, field in Config.get_fields().items():
        if field.validated_ty.ty is str or field.validated_ty.ty is Pattern:
            group.add_argument(
                f"--{field_name}",
                help=field.description,
            )
        elif field.validated_ty.ty is list:
            # NB: not currently settable via CLI
            pass
        else:
            raise AssertionError("unhandled type in Config")


def parse_config_from_cli(args: Namespace) -> Config:
    config = Config()

    if args.config:
        config = Config.parse_config(path=args.config)

    for field_name, field in Config.get_fields().items():
        args_val = getattr(args, field_name, None)
        if args_val:
            if field.validated_ty.ty is Pattern:
                config.set(field_name, args_val)
            elif field.validated_ty.ty is str:
                config.set(field_name, args_val)
            else:
                raise AssertionError("unhandled type in Config")

    return config



from pathlib import Path
from shutil import copy2

from ..config import Config, ValidatedConfig
from ..format import format_pyi_tree
from ..fs import Kind, Walker, remove
from .stubber import stubgen_single_file_src


def generate_stubs(
    input_path: Path,
    output_path: Path,
    config: Config | None = None,
):
    """
    Generate type stubs for all files found in `inp_root`,
    and writing them with the same structure to `out_root`.

    Optionally runs a formatter over the result.

    Optionally uses the provided `config` (otherwise the default config is used).
    """

    # NB: We want to support merging namespace module directories, so
    # we actually descent down into all non-namespace modules when processing the input directory
    inp: Path = input_path.resolve()
    out: Path = output_path.resolve()

    validated = (config or Config()).validate()

    walker = Walker(inp, out)

    _generate_stubs(walker, validated)

    format_pyi_tree(walker, validated)


def _generate_stubs(walker: Walker, config: ValidatedConfig):
    # pylint: disable=duplicate-code

    for event in walker.walk():
        inp = event.inp_path
        out = event.out_path

        match event.kind:
            case Kind.ROOT:
                print(f"Clean {event.out_rel_pattern}")
                remove(out)
            case Kind.COPY:
                print(f"Copy {event.out_rel_pattern}")
                out.parent.mkdir(parents=True, exist_ok=True)
                copy2(inp, out)
            case Kind.STUB:
                print(f"Stub {event.inp_rel_pattern} -> {event.out_rel_pattern}")
                content = inp.read_text()
                content = stubgen_single_file_src(content, Path(event.inp_rel_pattern), config)
                out.parent.mkdir(parents=True, exist_ok=True)
                out.write_text(content)



import subprocess
from pathlib import Path
from shutil import copy2
from tempfile import TemporaryDirectory

from ..config import ValidatedConfig
from ..format import format_pyi_tree
from ..fs import Event, Kind, Walker, remove, walk_dir


def generate_copy(inp: Path, out: Path):
    # in copy mode we just copy the pyi file as-is

    out.parent.mkdir(exist_ok=True, parents=True)
    copy2(inp, out)


def run_mypy(tmp: Path, dir_or_file: Path, out_dir: Path):
    cmd = [
        "stubgen",
        "--verbose",
        # "--inspect-mode",
        "--include-docstrings",
        "-o",
        str(out_dir),
        str(dir_or_file),
    ]
    cwd = tmp
    subprocess.run(
        cmd,
        cwd=cwd,
        check=True,
    )


def find_module_structure(path: Path) -> list[tuple[str, ...]]:
    modules: set[tuple[str, ...]] = set()

    for subpath in walk_dir(path):
        subpath_is_file = subpath.is_file()

        if subpath_is_file and subpath.suffix in (".py", ".pyi"):
            if subpath == path:
                modules.add(tuple())
            else:
                rel_parts = subpath.parent.relative_to(path).parts
                if subpath.stem == "__init__":
                    modules.add(tuple([*rel_parts]))
                else:
                    modules.add(tuple([*rel_parts, subpath.stem]))

    for found in modules.copy():
        while len(found) > 0:
            found = tuple(list(found)[:-1])
            modules.add(found)

    return sorted(modules)


def evaluate_structure(expected: list[tuple[str, ...]], current: list[tuple[str, ...]]) -> tuple[float, int]:
    """
    Evaluates how much the `current` structure matches the `expected` structure.

    Return two number:
    - a float that contains a value between 0.0 and 1.0 to express how much of the expected structure exists in the current structure.
    - a integer that counts the number of extra elements in the current structure that do not exist in the expected structure.
    """

    found = 0
    extra = 0

    for exp in expected:
        if exp in current:
            found += 1
    for ext in current:
        if ext not in expected:
            extra += 1

    # NB: If we have zero paths, we just compute a value of 0.0
    found_percent = float(found) / float(max(len(expected), 1))

    return (found_percent, extra)


def evaluate_structures(
    expected: list[tuple[str, ...]],
    current: list[tuple[str, ...]],
    ctx: tuple[str, ...],
) -> list[tuple[float, int, tuple[str, ...]]]:
    ret: list[tuple[float, int, tuple[str, ...]]] = []

    found_percent, extra = evaluate_structure(expected, current)
    ret.append((found_percent, extra, ctx))

    child_names: set[str] = set()
    for cur in current:
        if len(cur) < 1:
            continue
        child_names.add(cur[0])

    for child_name in child_names:
        children: set[tuple[str, ...]] = set()
        for cur in current:
            if len(cur) > 0 and cur[0] == child_name:
                children.add(cur[1:])
        child_ctx = ctx + (child_name,)
        ret.extend(evaluate_structures(expected, list(children), child_ctx))

    return ret


def find_mypy_out_subdir(mypy_inp_path: Path, mypy_out_path: Path) -> Path:
    """
    Given an input path to mypy, and the output directory that mypy wrote to,
    find the relative path in `mypy_out_path` that matches the directory nesting level
    at `mypy_inp_path`.

    This is needed because mypy seems to walk up in the directory tree from its
    input to discover the outermost python module, and include it in its
    entirety in the output directory tree, which makes it hard to match input
    and output 1 to 1.

    This function performs a search based on heuristics, as a python module
    could in theory contain multiple nested module structures that match the input
    path. We expect that this is not the case for most real-life code, however.
    """

    expected = find_module_structure(mypy_inp_path)
    current = find_module_structure(mypy_out_path)
    evaluated = evaluate_structures(expected, current, ())
    weighted = sorted(evaluated, key=lambda tup: (1.0 - tup[0], tup[1]))

    return Path(*weighted[0][2])


def generate_mypy(root: Event, stub_events: list[Event]):
    inp = root.inp_path

    # otherwise we invoke pyright with the right base directory to do a src import from
    with TemporaryDirectory() as tmpdir:
        tmp = Path(tmpdir)

        tmp_out = tmp / "out"
        tmp_out.mkdir()

        # call mypy
        run_mypy(tmp, inp, tmp_out)

        # Analyze mypy's output directory to figure out which of the files
        # in there match the input and output paths from the stub events.
        #
        # This is needed because mypy will walk up the directory tree for
        # its input until it finds something it identifies as the root of
        # the module tree, and then it will output the stubbed files under this
        # discovered directory path in the output directory.
        #
        # Additionally, mypy also will also not always match the kind of
        # module file of the source (`<name>/__init__.pyi` vs `<name>.pyi`).
        #
        # For example, stubbing `foo/bar/baz.py` _should_ just generate `baz.pyi`,
        # but can also end up generating `foo/bar/baz.pyi` or `foo/bar/baz/__init__.pyi`.
        #
        # We solve this issue heuristically in multiple phases:
        # - First we identify which output paths should be generated relative
        #   to the current root stubbing Event.
        # - Then we heuristically try to figure out at which nesting level in the
        #   mypy output directory we find the type stubs that match the same level
        #   as the path argument passed to mypy.
        # - Finally we combine the both, by looking at the stub files at the
        #   right location in the mypy output, and moving them to the the
        #   expected output paths.
        root_relative_out_paths = get_root_relative_out_paths(root, stub_events)
        mypy_out_subdir = find_mypy_out_subdir(inp, tmp_out)
        for stub_out_root_relative, stub_out in root_relative_out_paths:
            # Compute the expected path to the module, in form of a directory
            # path whose last component contains the name of the module.
            #
            # This is used as a placeholder to then derive the two possible
            # module file paths from it.
            mypy_out_mod_path = tmp_out / mypy_out_subdir / stub_out_root_relative
            if mypy_out_mod_path.stem == "__init__":
                mypy_out_mod_path = mypy_out_mod_path.parent
            else:
                mypy_out_mod_path = mypy_out_mod_path.with_suffix("")
            assert mypy_out_mod_path.suffix == ""

            # NB: Mypy might generate either `<name>/__init__.pyi` or `<name>.pyi`,
            # so we check both variants
            mod_file_candidates = (
                mypy_out_mod_path.with_suffix(".pyi"),
                mypy_out_mod_path / "__init__.pyi",
            )
            for candidate in mod_file_candidates:
                if candidate.exists():
                    stub_out.parent.mkdir(exist_ok=True, parents=True)
                    candidate.rename(stub_out)
                    break


def group_per_root(walker: Walker) -> list[tuple[Event, list[Event], list[Event]]]:
    """
    Group up Walker Events, such that each ROOT event is matched with
    all its child STUB and COPY events.
    """

    roots: list[tuple[Event, list[Event], list[Event]]] = []

    for event in walker.walk():
        match event.kind:
            case Kind.ROOT:
                roots.append((event, [], []))
            case Kind.STUB:
                root, stub_events, _ = roots[-1]
                if root.is_file:
                    assert event.inp_path == root.inp_path
                else:
                    assert root.inp_path in [event.inp_path, *event.inp_path.parents]
                stub_events.append(event)
            case Kind.COPY:
                root, _, copy_events = roots[-1]
                if root.is_file:
                    assert event.inp_path == root.inp_path
                else:
                    assert root.inp_path in [event.inp_path, *event.inp_path.parents]
                copy_events.append(event)

    return roots


def get_root_relative_out_paths(root: Event, stub_events: list[Event]) -> list[tuple[Path, Path]]:
    """Returns a list of tuples that contain the path of a stub file both relative to its root, and absolute"""

    stub_out_paths: list[tuple[Path, Path]] = []
    for stub_event in stub_events:
        stub_out_paths.append((stub_event.out_path.relative_to(root.out_path), stub_event.out_path))

    return stub_out_paths


def generate(inp_root: Path, out_root: Path, config: ValidatedConfig):
    walker = Walker(inp_root, out_root)
    roots = group_per_root(walker)
    for root, stub_events, copy_events in roots:
        print(f"Clean {root.out_rel_pattern}")
        remove(root.out_path)

        # only call mypy if there is at least one file we need to have stubbed
        if stub_events and (not root.is_file or (root.inp_path.suffix in (".py", ".pyi"))):
            print(f"Stub {root.inp_rel_pattern} -> {root.out_rel_pattern}")
            generate_mypy(root, stub_events)

        # Files that can be kept as-is get copied directly afterwards. This
        # also ensures any .pyi files mypy might have copied over will be ignored
        # in favour of their original content.
        for event in copy_events:
            print(f"Copy {event.out_rel_pattern}")
            generate_copy(event.inp_path, event.out_path)

    format_pyi_tree(walker, config)



from tomli_w import dumps as dumps  # pylint: disable=unused-import

try:
    from tomli import loads as loads  # pylint: disable=unused-import
except ImportError:
    from tomllib import loads as loads  # type: ignore



import os
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from shutil import rmtree
from typing import Iterable


def walk_dir(root: Path) -> Iterable[Path]:
    assert root.exists(), f"Path does not exist: >{root}<"

    if root.is_file():
        yield Path(root)

    for dirname, _, filenames in os.walk(root):
        yield Path(dirname)
        for filename in sorted(filenames):
            yield Path(dirname) / filename


def remove(path: Path):
    if path.is_file():
        path.unlink()
    elif path.is_dir():
        rmtree(path)


def find_module_roots(path: Path) -> Iterable[Path]:
    """
    Find all paths that define the root of a module tree for the purpose of type
    checking.

    This includes isolated `.py` and `.pyi` files, as well as directories that contain
    `__init__.py` files.

    The paths will not overlap, so if a directory is yield, no child files from it will
    be yielded.
    """

    # the entire directory tree defines a module
    if path.is_dir() and ((path / "__init__.py").exists() or (path / "__init__.pyi").exists()):
        yield path
        return

    # this single file defines a module
    if path.is_file() and (path.suffix in (".py", ".pyi") or path.name == "py.typed"):
        yield path
        return

    # recursively look in children instead
    if path.is_dir():
        for child in sorted(path.iterdir()):
            yield from find_module_roots(child)


class Kind(Enum):
    ROOT = auto()
    COPY = auto()
    STUB = auto()


@dataclass
class Event:
    walker: "Walker"

    inp_path: Path
    inp_rel_pattern: str

    out_path: Path
    out_rel_pattern: str

    kind: Kind
    is_file: bool


# pylint: disable-next=too-few-public-methods
class Walker:
    inp_root: Path
    out_root: Path

    root_is_file: bool

    cache: list[Event] | None

    def __init__(self, inp_root: Path, out_root: Path):
        assert inp_root.exists()

        self.inp_root = inp_root.resolve()
        self.out_root = out_root.resolve()

        self.root_is_file = self.inp_root.is_file()

        # support writing a single file to an output directory
        if self.root_is_file and self.out_root.is_dir():
            self.out_root = self.out_root / self.inp_root.name
            if self.inp_root.suffix == ".py":
                self.out_root = self.out_root.with_suffix(".pyi")

        self.cache = None

    def _get_paths(self, inp_path: Path) -> tuple[Path, str, str]:
        """
        Given an input path:
        - Return the matching output path
        - Return the matching input pattern string
        - Return the matching output pattern string
        """

        # Relative path for input file
        rel_inp = inp_path.relative_to(self.inp_root)

        # Output path with potentially wrong target file name
        out_path = self.out_root / rel_inp

        if self.root_is_file:
            # If root is a file, we keep the output as it is
            inp_rel_pattern = inp_path.name
            out_rel_pattern = out_path.name
        else:
            # Otherwise we determine the new file suffixes
            if inp_path.suffix == ".py":
                out_path = out_path.with_suffix(".pyi")
            inp_rel_pattern = str(rel_inp)
            out_rel_pattern = str(out_path.relative_to(self.out_root))

        return (out_path, inp_rel_pattern, out_rel_pattern)

    def _walk_module_root(self, inp_path_root: Path) -> Iterable[Event]:
        """Yield all walk events relative to a python module root."""

        for inp_path in walk_dir(inp_path_root):
            out_path, inp_rel_pattern, out_rel_pattern = self._get_paths(inp_path)

            # NB: If we have both a .py and pyi file, we prefer keeping the .pyi file
            if inp_path.suffix == ".py" and not inp_path.with_suffix(".pyi").exists():
                # this file we stub
                kind = Kind.STUB
            elif inp_path.suffix == ".pyi" or inp_path.name == "py.typed":
                # these files we just copy
                kind = Kind.COPY
            else:
                # other files we silently ignore
                continue

            yield Event(
                walker=self,
                inp_path=inp_path,
                out_path=out_path,
                inp_rel_pattern=inp_rel_pattern,
                out_rel_pattern=out_rel_pattern,
                kind=kind,
                is_file=not inp_path.is_dir(),  # NB: This also treats special files as files
            )

    def _walk(self) -> Iterable[Event]:
        """Yield all walk events."""

        for inp_path in find_module_roots(self.inp_root):
            out_path, inp_rel_pattern, out_rel_pattern = self._get_paths(inp_path)

            # we want to clean the entire module root
            yield Event(
                walker=self,
                inp_path=inp_path,
                out_path=out_path,
                inp_rel_pattern=inp_rel_pattern,
                out_rel_pattern=out_rel_pattern,
                kind=Kind.ROOT,
                is_file=not inp_path.is_dir(),  # NB: This also treats special files as files
            )
            yield from self._walk_module_root(inp_path)

    def walk(self) -> Iterable[Event]:
        """Yield all walk events."""

        if self.cache is not None:
            yield from self.cache
        else:
            cache: list[Event] = []
            for event in self._walk():
                yield event
                cache.append(event)
            self.cache = cache
